
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sched.h>
#include <assert.h>

#define DEBUG(s) {int result; result = (s); \
       assert(result == 0);}

#define N 5

#define NB_PRODUCER 3
#define NB_CONSUMER 2
#define NB_THREADS (NB_PRODUCER + NB_CONSUMER)


/* Shared buffer */
char shared_buffer[N];

/* Number of items */
/* Note the volatile class! */
volatile int nb_produced = 0, nb_consumed = 0;

/* Condition variables */
pthread_cond_t not_full_cond;
pthread_cond_t not_empty_cond;

pthread_mutex_t mp;

void produce_value(char c, int n) 
{
  DEBUG(pthread_mutex_lock(&mp));

  while ((nb_produced - nb_consumed) == N) 
    DEBUG(pthread_cond_wait(&not_full_cond, &mp));

  printf("%d: Producing '%c'...\n", n, c);
  nb_produced++;
  shared_buffer[(nb_produced - 1)%N] = c;
  printf("%d: Produced...\n", n);

  DEBUG(pthread_cond_broadcast(&not_empty_cond));

  DEBUG(pthread_mutex_unlock(&mp));
}

char consume_value(int n)
{
  char c;
  DEBUG(pthread_mutex_lock(&mp));

  while ((nb_produced - nb_consumed) == 0) 
    DEBUG(pthread_cond_wait(&not_empty_cond, &mp));

  printf("\t\t%d: Consuming...\n", n);
  nb_consumed++;
  c  = shared_buffer[(nb_consumed - 1)%N];
  printf("\t\t%d: Consumed '%c'\n", n ,c);

  DEBUG(pthread_cond_broadcast(&not_full_cond));

  DEBUG(pthread_mutex_unlock(&mp));
  return c;
}


void *producer_func (void *s)
{
  int n = (int) s;
  int i;
  for (i = 0; ; (i == 25)?(i = 0):(i++)) {
    char c;
    c = (char) ('a'+i);
    produce_value(c, n);
    sched_yield();
  }
}

void *consumer_func (void *s)
{
  int n = (int) s;
  char c;
  for (;;) {
    c = consume_value(n);
    sched_yield();
  }
}

int main(void)
{ 
  void *status;
  pthread_t thread_tid[NB_THREADS];
  pthread_attr_t attributes;
  int i;

  DEBUG(pthread_mutex_init(&mp, NULL));

  DEBUG(pthread_cond_init(&not_empty_cond, NULL));
  DEBUG(pthread_cond_init(&not_full_cond, NULL));

  DEBUG(pthread_attr_init(&attributes));

  for (i = 0; i < NB_PRODUCER; i++) {
    DEBUG(pthread_create(&(thread_tid[i]), &attributes, 
			 &producer_func, (void *)(i)));
  }
  
  for (i = 0; i < NB_CONSUMER; i++) {
    DEBUG(pthread_create(&(thread_tid[i]), &attributes, 
			 &consumer_func, (void *)(i)));
  }
  
  for (i = 0; i < NB_THREADS; i++) {
    DEBUG(pthread_join(thread_tid[i], &status));
  }

  DEBUG(pthread_cond_destroy(&not_empty_cond));
  DEBUG(pthread_cond_destroy(&not_full_cond));

  DEBUG(pthread_mutex_destroy(&mp));

  return 0;
}


